$TITLE MultiCommodityFlowStartPoint OA4202 2010
$INLINECOM { }
 OPTIONS
   SOLPRINT =     OFF,
   DECIMALS =       1,
   LIMCOL   =      10,
   LIMROW   =      10,
   RESLIM   =      60, {max seconds}
   ITERLIM  =99999999, {max iterations}
   LP       =     CPLEX,
   MIP      =     CPLEX
 ;


$offlisting

$ontext
============================================================================
With the following few lines of code, we are reading in the network specification.
- We read in the set of nodes, from the nodes.csv file
- We read in the set of all arcs, from the arcs_set.csv file
- We define the subset of arcs which we can interdict
- Finally, we read in the arc data, from the arcs_data.csv file
============================================================================
$offtext


SET
    n   nodes  /
$INCLUDE  sample_nodes.csv
 /;

alias(n,i,j)   {(i,j) is for arc distance and flow}

SET
    k   commodities  /
$INCLUDE  sample_commodities.csv
 /;

SET
   arcs(i,j)         arclist  /
$ONDELIM
$INCLUDE  sample_arcs_set.csv
$OFFDELIM
 /;

SET
   attackarcs(i,j)         the arcs on which we can place interdictions  /
$ONDELIM
$INCLUDE  sample_attack_arcs_set.csv
$OFFDELIM
 /;

table nodedata(i,k,*)
$ONDELIM
$INCLUDE  sample_nodes_data.csv
$OFFDELIM
;

table arcdata(i,j,*)
$ONDELIM
$INCLUDE  sample_arcs_data.csv
$OFFDELIM
;

table arccommoditydata(i,j,k,*)
$ONDELIM
$INCLUDE  sample_arcs_commodity_data.csv
$OFFDELIM
;


$ontext
============================================================================
We'll need the nC calculation to eliminate arcs completely
============================================================================
$offtext


SCALARS nC ;
 nC=0.0 ;

LOOP(k,
 LOOP(arcs(i,j),
   IF( nC<ABS(arccommoditydata(i,j,k,'Cost')),
       nC=ABS(arccommoditydata(i,j,k,'Cost')) ;
   );
 );
);

nC = nC * card(n);

display nC;


$ontext
============================================================================
Lets define the primal multicommodity flow LP.

Our definition includes the parameters xbar(i,j).  We'll use them later, to
re-solve for the multicommodity flow, given an interdiction plan X.  For now, we'll just
set the xbar(i,j) to zero.
============================================================================
$offtext

 PARAMETERS
   delay
   xbar(i,j)
 ;

 delay = nC;

 LOOP(arcs(i,j),
   xbar(i,j)=0 ;
 );

 VARIABLE
   Zprimal           objective function value
 ;

 POSITIVE VARIABLES
   Y(i,j,k)            flow on arc i j for commodity k
 ;
 
 POSITIVE VARIABLES
   UnsatisfiedDemand(i,k)            SupplyDemand that could not be moved at node i for commodity k
 ;

 EQUATIONS
   OBJECTIVE
   NETFLOW(n,k)
   CAPACITY(i,j)
 ;

 OBJECTIVE..
   Zprimal =e=  SUM(k, SUM(arcs(i,j),(  arccommoditydata(i,j,k,'Cost') + delay*xbar(i,j)$(attackarcs(i,j) ) ) *Y(i,j,k) ) ) + nC/2*SUM(k, SUM(j,UnsatisfiedDemand(j,k)) )
;

 NETFLOW(j,k)..
   SUM(arcs(i,j),Y(i,j,k)) - SUM(arcs(j,i),Y(j,i,k))      {+ flow in - flow out}
     =e= nodedata(j,k,'SupplyDemand') + UnsatisfiedDemand(j,k)$(nodedata(j,k,'SupplyDemand') lt 0)
                                      - UnsatisfiedDemand(j,k)$(nodedata(j,k,'SupplyDemand') ge 0)
 ;
 
 CAPACITY(arcs(i,j))..
   SUM(k, Y(i,j,k)) =l= arcdata(i,j,'Capacity')
 ;

 MODEL MultiCommodityFlow
 /
   OBJECTIVE
   NETFLOW
   CAPACITY
 /;

$ontext
============================================================================
Now, lets solve for the multicommodity flow when there are no interdictions.
============================================================================
$offtext

 SOLVE MultiCommodityFlow USING LP MINIMIZING Zprimal;       { find the uninterdicted multicommodity flow}

$ontext
============================================================================
Lets print the uninterdicted multicommodity flow solution to a file called modelOutput.out
============================================================================
$offtext


 file out /modelOutput.out/;
 put out;

 put 'nC is', nC / ;
 put '****' / ;
 put 'solve MultiCommodityFlow with no interdictions' / ;
 LOOP(k,
     LOOP(arcs(i,j)$(Y.l(i,j,k)>0),
       put '    ', i.tl,' -> ', j.tl, ' ... ', Y.l(i,j,k) , ' ', k.tl / ;
     );
     LOOP(j$(UnsatisfiedDemand.l(j,k)>0),
       put '   SupplyDemand that could not be moved on node ', j.tl, ' for commodity ', k.tl, ' is ', (sign(nodedata(j,k,'SupplyDemand')) * UnsatisfiedDemand.l(j,k)) / ;
     );
 );
 put 'total cost =',Zprimal.l:10:1 / ;


$ontext
============================================================================
Now, lets define the MIP for computing interdictions.

We'll introduce a scalar, 'nAttacks', that specifies the number of
interdictions to compute.

We want to maximize the multicommodity flow.  So, the problem we want to solve is a
'max min' problem.  To put this into the solver, we'll take the dual of the
inner problem, so the problem becomes a 'max max'.  That is why in the
following code, you'll see the dual of the multicommodity flow problem.

The binary variables X(i,j), will specify whether we attack arc i j.
============================================================================
$offtext


SCALARS nAttacks;

nAttacks = 1;

 VARIABLES
    Zdual
 ;

 VARIABLES
   Rho(n,k)              dual variables for the netflow constraints
 ;

 NEGATIVE VARIABLES
   Pi(i,j)              dual variables for the capacity constraints
 ;

 BINARY VARIABLES
   X(i,j)            should we attack arc i j
 ;

 EQUATIONS
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS(i,j,k)
   DUAL_CONSTRAINTS_UNSAT(j,k)
   ATTACK_LIMIT
 ;

 DUAL_OBJECTIVE..
   Zdual =e= SUM(arcs(i,j), arcdata(i,j,'Capacity')*Pi(i,j)) + SUM(k, SUM(j, nodedata(j,k,'SupplyDemand')*Rho(j,k)))
 ;

 DUAL_CONSTRAINTS(i,j,k)$(arcs(i,j))..
   Rho(j,k)-Rho(i,k)+Pi(i,j) =l= arccommoditydata(i,j,k,'Cost') + delay*X(i,j)$(attackarcs(i,j) )
 ;

 DUAL_CONSTRAINTS_UNSAT(j,k)..
 -Rho(j,k)$(nodedata(j,'SupplyDemand') lt 0) + Rho(j,k)$(nodedata(j,'SupplyDemand') ge 0) =l= nC/2
 ;

 ATTACK_LIMIT..
   SUM(attackarcs(i,j),X(i,j)) =l= nAttacks
 ;
 
 MODEL InterdictionMIP
 /
   DUAL_OBJECTIVE
   DUAL_CONSTRAINTS
   DUAL_CONSTRAINTS_UNSAT
   ATTACK_LIMIT
 /;

$ontext
============================================================================
Now, lets solve for the best interdiction plan.
============================================================================
$offtext

 InterdictionMIP.optfile = 1;

 SOLVE InterdictionMIP USING MIP MAXIMIZING Zdual;       { Find the best interdiction plan }

$ontext
============================================================================
Lets print the interdiction plan into the output file we had opened, modelOutput.out
============================================================================
$offtext

display X.l;

 put '****' / ;
 put 'interdiction plan with ', nAttacks, ' nAttacks:' / ;
 LOOP(arcs(i,j)$(X.l(i,j)>0),
   put '   attack arc: ',i.tl,' -> ', j.tl / ;
 );
 put 'cost with interdictions in place =',Zdual.l:10:1 / ;

$ontext
============================================================================
Finally, now that we have computed attack locations, lets re-solve for 
the operator's new best flow.
============================================================================
$offtext

 LOOP(arcs(i,j),                        { transfer the interdiction solution to constant parameter }
   xbar(i,j)=X.l(i,j) ;
 );

 SOLVE MultiCommodityFlow USING LP MINIMIZING Zprimal;       { find multicommodity flow plan with the interdictions in place }

 put '****' / ;
 put 'solve MultiCommodityFlow with interdictions in place' / ;
 put 'operator best response:' / ;
 LOOP(k,
     LOOP(arcs(i,j)$(Y.l(i,j,k)>0),
       put '    ', i.tl,' -> ', j.tl, ' ... ', Y.l(i,j,k) , ' ', k.tl / ;
     );
     LOOP(j$(UnsatisfiedDemand.l(j,k)>0),
       put '   SupplyDemand that could not be moved on node ', j.tl, ' for commodity ', k.tl, ' is ', (sign(nodedata(j,k,'SupplyDemand')) * UnsatisfiedDemand.l(j,k)) / ;
     );
 );
 put 'total cost =',Zprimal.l:10:1 / ;
